//***********************************************************************************************************
//
// The X_Tokenizer class takes in a delimited string and parses it into tokens
//
//
// author	Tony Madden	(tm@mqware.com)
//***********************************************************************************************************
#include	<x_tokenizer.hpp>

X_Tokenizer::X_Tokenizer(){
	initialize();
	;
}

//********************************************************************
// Parse the string using spaces as the delimiter
// NOTE - special case. Any number of consecutive spaces counts as one delimiter (but see next reset() method)
//
// arg_sDelimited			delimited data
//********************************************************************
void X_Tokenizer::parse(const string& arg_sDelimited){
	reset();
	tokenParser(arg_sDelimited,' ',true);
}

//********************************************************************
// Parse the string using arg_cDelim as the delimiter
// NOTE - If arg_cDelim is a space, then each one of consecutive spaces counts as a separate delimiter
//
// arg_sDelimited			delimited data
// arg_cDelim				the delimiter
//********************************************************************
void X_Tokenizer::parse(const string& arg_sDelimited,char arg_cDelim){
	reset();
	tokenParser(arg_sDelimited,arg_cDelim,false);
}

//********************************************************************
// Parse the string using separate start and end delimiters
// For example, if the string is " ELEMENT={abcdef}" ,
//	to isolate "abcdef", arg_sStartDelim would be "ELEMENT={" (or a substring thereof) and arsEndDelim would be "}"
//
// This function is useful for extracting values from MQSeries MQSC output, for example.
//
// arg_sDelimited			delimited data
// arg_sStartDelim			start delimiter
// arg_sEndDelim			end delimiter
//********************************************************************
void X_Tokenizer::parse(const string& arg_sDelimited,string arg_sStartDelim,string arg_sEndDelim){
	reset();
	enclosedByBothTokens(arg_sDelimited,arg_sStartDelim,arg_sEndDelim);
}

//********************************************************************
// Reset - performed each time a parse(...) method is called
//********************************************************************
void	X_Tokenizer::reset(){
	tokenVector.resize(0);
	lCurrentIndex=0;
	
	
}
//********************************************************************
// Initialize - performed at object instantiation only
//********************************************************************
void	X_Tokenizer::initialize(){
	cEscapeCharacter = TK_DEFAULT_ESCAPE_CHARACTER;
	bProcessEscapeCharacters = false;
	bTrimTokens = true;
	
}
//********************************************************************
// Set escape character
//********************************************************************
void	X_Tokenizer::setEscapeCharacter(char	arg_c){
	cEscapeCharacter = arg_c;
}
//********************************************************************
// Enable escape character checking (by default it is set off (flase)
//********************************************************************
void	X_Tokenizer::processEscapeCharacters(bool arg_b){
	bProcessEscapeCharacters = arg_b;
}
//********************************************************************
//
//********************************************************************
bool	X_Tokenizer::isProcessEscapeCharacters(){
	return bProcessEscapeCharacters;
}
//********************************************************************
// 
//********************************************************************
char	X_Tokenizer::getEscapeCharacter(){
	return cEscapeCharacter;
}

//********************************************************************
// Set internal index
//********************************************************************
bool	X_Tokenizer::setCurrentIndex(long arg_Index){
	bool	rc;
	lCurrentIndex=0;

	if (tokenVector.size() > arg_Index){
		lCurrentIndex = arg_Index;
		rc = true;
	}else{
		rc = false;
	}
	return rc;
}

//********************************************************************
// Get address of the vector which holds all tokens
//********************************************************************
const vector<string>*	X_Tokenizer::getAllTokens(){
	return &tokenVector;
}

//********************************************************************
// Load copy of the vector which holds all tokens
//********************************************************************
void	X_Tokenizer::getAllTokens(vector<string>* arg_Vector){
	*arg_Vector =  tokenVector;
}

//********************************************************************
// Get token at the specified index (i.e. offset)
//
// return	true if token found, false if no token found
//********************************************************************
bool			X_Tokenizer::getToken(long arg_lInd,string* arg_psToken){
	bool	bRc = false;

	if (tokenVector.size() > arg_lInd){
		*arg_psToken = tokenVector[arg_lInd];
		bRc = true;
	}else{
		*arg_psToken = "";
		bRc = false;
	}
	return bRc;
}


//********************************************************************
// Get token at the specified index (i.e. offset)
//
// return	the token if found, return zero length string if not found
//********************************************************************
string			X_Tokenizer::getToken(long arg_lInd){
	string	str;

	getToken(arg_lInd,&str);

	return str;
}

//********************************************************************
// Get first token 
//
// return	 the token if found, return zero length string if not found
//********************************************************************
string			X_Tokenizer::getFirstToken(){
	return getToken((long)0);
}

//********************************************************************
// Get first token
//
// return	 true if token found, false if no token found
//********************************************************************
bool			X_Tokenizer::getFirstToken(string* arg_psToken){
	return getToken((long)0,arg_psToken);
}

//********************************************************************
// Get next token
//
// return	 true if token found, false if no token found
//********************************************************************
bool			X_Tokenizer::getNextToken(string* arg_psToken){
	return getToken(lCurrentIndex++,arg_psToken);
}

//********************************************************************
// get next token 
//
// return the token if found, return zero length string if not found
//********************************************************************
string			X_Tokenizer::getNextToken(){
	return getToken(lCurrentIndex++);
}
string	X_Tokenizer::getOriginalString(){
	return sOriginalString;
}

//********************************************************************
// Get the number of tokens
//********************************************************************
long	X_Tokenizer::numberOfTokens(){
	return tokenVector.size();
}

//********************************************************************
// Parse the delimited string 
//
// arg_sDelimited							the delimited data
// arg_cDelim								the delimiter
// arg_bConsecutiveSpacesAreOneDelimiter	if true then treat consecutive spaces between two tokens as a single delimiter
//											if false then treat each of consecutive spaces between two values as separate delimiters
// arg_pTokenVector							vector to be loaded with tokens
//********************************************************************
void X_Tokenizer::tokenParser(const string&		arg_sDelimited,
							  char				arg_cDelim,
							  bool				arg_bConsecutiveSpacesAreOneDelimiter){

	string::size_type stStart = 0;
	bool bSpaceDelimitedAndConsecutiveSpacesAreOneDelimiter ;

	sOriginalString = arg_sDelimited;

	tokenVector.clear();

	if (arg_cDelim == ' ' && arg_bConsecutiveSpacesAreOneDelimiter){
		bSpaceDelimitedAndConsecutiveSpacesAreOneDelimiter = true;
	}else{
		bSpaceDelimitedAndConsecutiveSpacesAreOneDelimiter = false;
	}
			
	bool	bEscapedDelimiter ;
	for (string::size_type st = 0;;st++){
		bEscapedDelimiter = false;
		stStart = st;
		if (bSpaceDelimitedAndConsecutiveSpacesAreOneDelimiter){
			st = arg_sDelimited.find_first_not_of(arg_cDelim,st);
			if (st == string::npos){
				break;
			}
		}
		// Find a delimiter
		st = arg_sDelimited.find(arg_cDelim,st);
		if (st == string::npos){
			// No more delimiters so add entry to vector
			if (bTrimTokens){
				string s = arg_sDelimited.substr(stStart);
				s = X_String::trim(s);
				tokenVector.push_back(s);
			}else{
				tokenVector.push_back(arg_sDelimited.substr(stStart));
			}
			break;
		}else{
			string sElement;
			if (bProcessEscapeCharacters && st > 0){
				// Process escape characters
				if (arg_sDelimited.substr(st-1,1)[0] == cEscapeCharacter){
					// The delimiter itself is escaped, so treat it as data
					if (st+1 >= arg_sDelimited.size()){
						// delimiter is final byte, so create delimited element (which will have the delimiter as final byte)
						sElement = X_String::removeEscapeCharacter(arg_sDelimited.substr(stStart,st - stStart + 1),cEscapeCharacter);
						bEscapedDelimiter = true;
					}else{
						// delimiter is just data, so iterate to find the next "real" delimiter
						continue;
					}
				}else{
					// Remove escape characters within the delimited string
					sElement = X_String::removeEscapeCharacter(arg_sDelimited.substr(stStart,st - stStart),cEscapeCharacter);
				}
			}else{
				// ignore escape characters
				sElement = arg_sDelimited.substr(stStart,st - stStart);
			}

			if (bTrimTokens){
				// remove leading/trailing spaces
				tokenVector.push_back(X_String::trim(sElement));
			}else{
				tokenVector.push_back(sElement);
			}
			
			if (bSpaceDelimitedAndConsecutiveSpacesAreOneDelimiter){
				stStart = arg_sDelimited.find_first_not_of(arg_cDelim,st+1);
				if (stStart == string::npos){
					break;
				}
			}
			if (st == arg_sDelimited.size() - 1){
				if (!bEscapedDelimiter){
					// the final byte is a delimiter. This means that the final delimited value is a null string.
					tokenVector.push_back(string(""));
				}
				break;
			}
		}
	}

}

//********************************************************************
// Parse the string using separate start and end delimiters
// For example, if the string is " ELEMENT={abcdef}" , then in order to isolate "abcdef", arg_sStartDelim would 
// be "ELEMENT={" (or a substring thereof) and arg_sEndDelim would be "}"
//
// This function is useful for extracting values from MQSeries MQSC output.
//
// arg_sDelimited			delimited data
// arg_sStartDelim			start delimiter
// arg_sEndDelim			end delimiter
//********************************************************************
void	X_Tokenizer::enclosedByBothTokens(const string& arg_Delimited,
												  const string& arg_sStartDelim,
												  const string& arg_sEndDelim){
	string::size_type	stStart;
	string::size_type	stEnd;

	stStart = 0;

	while (true){

		stStart = arg_Delimited.find(arg_sStartDelim,stStart);
		if (stStart == string::npos){
			break;
		}
		stStart += arg_sStartDelim.size();
		stEnd = arg_Delimited.find(arg_sEndDelim,stStart);
		if (stEnd == string::npos){
			break;
		}
		tokenVector.push_back(arg_Delimited.substr(stStart,stEnd-stStart));
	}


}

bool	X_Tokenizer::isTrimTokens(){
	return bTrimTokens;
}
void	X_Tokenizer::trimTokens(bool arg_b){
	bTrimTokens = arg_b;
}

string	X_Tokenizer::getVersion(){
	return X_TOKENIZER_VERSION;
}
